--
-- This module implements the core of the FACTS rule engine and a simplified
-- simulation of a sensor network for basic testing.
--

module FACTS.Main (

	Condition(Exists, Eval),
	Statement(Define, Retract, Send, Set, Call),
	Expression(Constant, UnaryOperation, BinaryOperation),
	ComparisonOperation(Equals, NotEquals, LessThan, GreaterThan),
	UnaryArithmeticOperation(Count, Sum, Product),
	BinaryArithmeticOperation(Add, Subtract, Multiply, Divide),

	broadcastMAC, minPriority, maxPriority,

	varBool, varInt, varFloat, varString, varSlot,
	newSimulation, newEvent, newNetwork, newNode,
	newName, newRuleset, newSlot, newSimpleSlot, newRule, newFact, newProperty, newFunction,

	runSimulation, addNodes, addNode, addRulesets, addRuleset,
	addRules, addRule, addFacts, addFact, addFunctions, addFunction

) where



--
-- General definitions.
--

type MAC = Int
type Position = (Float, Float)
type Name = String
type Key = String
type Pattern = String
-- Possible types of properties of facts.
data Type =
	TypeBool | TypeInt | TypeFloat | TypeString | TypeSlot
	deriving (Eq, Ord)
instance Show Type where
	show TypeBool = "bool"
	show TypeInt = "int"
	show TypeFloat = "float"
	show TypeString = "string"
	show TypeSlot = "slot"
-- Repeating the type in the value part of the key-value-pair (that makes up
-- a property of a fact) is some necessary overhead caused by Haskell syntax.
data Value =
	ValueBool Bool | ValueInt Int | ValueFloat Float | ValueString String | ValueSlot Slot
	deriving (Eq, Ord)
instance Show Value where
	show (ValueBool bool) = show bool
	show (ValueInt int) = show int
	show (ValueFloat float) = show float
	show (ValueString string) = string
	show (ValueSlot slot) = show slot
data Slot = Slot Name Key [Condition]
	deriving (Eq, Ord)
instance Show Slot where
	show (Slot name key conditions) = name ++ separator ++ key ++ conditionsString
		where	conditionsString =	if (null conditions) then ""
						else " (" ++ (join " && " (map show conditions)) ++ ")"
data Variable = Variable Type Value
	deriving (Eq, Ord)
instance Show Variable where
	show (Variable ttype value) = "(" ++ show ttype ++ ") " ++ show value

--
-- Simulation definitions.
--

type Time = Int
-- A simulation is completely defined by its initial state and the list of
-- events that happen while it is running.
data Simulation = Simulation State [Event]
	deriving (Show)
-- A state of the simulation consists of the simulation step count, the current
-- time, and most importantly the network of sensor nodes at that time.
data State = State Int Time Network
instance Show State where
	show (State step time (Network nodes)) =
		"Network with " ++ show (length nodes) ++ " node(s) at time " ++ show time ++ ":\n\n" ++ show (Network nodes)
-- An event is a list of statements that are applied to a particular node
-- (identified by its MAC) at a particular step in the simulation run. Applying
-- events at a particular simulated time is not supported.
data Event = Event Int MAC [Statement]
	deriving (Show)

--
-- Network definitions.
--

-- The network is a list of nodes.
data Network = Network [Node]
instance Show Network where
	show (Network nodes) = unlines (map show (reverse nodes))
-- A sensor node is a tuple of its MAC, its position, and of course lists of
-- rules, facts, and functions.
data Node = Node MAC Position [Rule] [Fact] [Function]
	deriving (Eq)
instance Show Node where
	show (Node mac (x,y) rules facts functions) =
		"Node #" ++ show mac ++ " @ (" ++ show x ++ ", " ++ show y ++ ")\n"
		++ show (Ruleset rules facts functions)
-- Similarly, a ruleset is a tuple of rules, facts, and functions.
data Ruleset = Ruleset [Rule] [Fact] [Function]
instance Show Ruleset where
	show (Ruleset rules facts functions) =
		"\t|-> Rules:" ++ rulesString
		++ "\t|-> Facts:" ++ factsString
		++ "\t|-> Functions:" ++ functionsString
		where	rulesString = if showRules then "\n" ++ unlines (map ((spacing ++).show) (reverse rules)) else "\t[not shown]\n"
			factsString = if showFacts then "\n" ++ unlines (map ((spacing ++).show) (reverse facts)) else "\t[not shown]\n"
			functionsString = if showFunctions then "\n" ++ unlines (map ((spacing ++).show) (reverse functions)) else "\t[not shown]\n"
			spacing = "\t|   "


--
-- Rule definitions.
--

type Priority = Int
-- A rule is a tuple of its name, priority and lists of conditions and
-- statements.
data Rule = Rule Name Priority [Condition] [Statement]
	deriving (Eq)
instance Show Rule where
	show (Rule name priority conditions statements) =
		"(" ++ show priority ++ ") " ++ name ++ conditionsString ++ statementsString
		where	conditionsString =	if (null conditions) then ""
						else " <- " ++ (join " && " (map show conditions))
			statementsString =	if (null statements) then ""
						else " -> " ++ (join "; " (map show statements))
data Condition =
	Exists Slot
	| Eval Expression ComparisonOperation Expression
	deriving (Eq, Ord)
instance Show Condition where
	show (Exists (Slot pattern _ conditions)) = pattern ++ conditionsString
		where	conditionsString =	if (null conditions) then ""
						else " (" ++ (join " && " (map show conditions)) ++ ")"
	show (Eval expr1 operation expr2) = show expr1 ++ " " ++ show operation ++ " " ++ show expr2
data Statement =
	Define Name [(Key, Expression)]
	| Retract Slot
	| Send Expression Expression Slot
	| Set Slot Expression
	| Call Function [Variable]
	deriving (Eq)
instance Show Statement where
	show (Define name keyExpressionPairs) = "Define (" ++ show name ++ keyExpressionPairsString ++ ")"
		where 	keyExpressionPairsString =	if (null keyExpressionPairs) then ""
							else ": " ++ (join ", " (map show keyExpressionPairs))
	show (Retract (Slot pattern _ conditions)) = "Retract " ++ pattern ++ conditionsString
		where	conditionsString =	if (null conditions) then ""
						else " (" ++ (join " && " (map show conditions)) ++ ")"
	show (Send receiverMACExpression txPowerExpression (Slot pattern _ conditions)) = "Send (" ++ show receiverMACExpression ++ " " ++ show txPowerExpression ++ " " ++ pattern ++ conditionsString ++ ")"
		where	conditionsString =	if (null conditions) then ""
						else " (" ++ (join " && " (map show conditions)) ++ ")"
	show (Set slot expression) = "Set (" ++ show slot ++ " = " ++ show expression ++ ")"
	show (Call function vars) = "Call (" ++ show function ++ ") with var(s) " ++ (join ", " (map show vars))
data Expression =
	Constant Variable
	| UnaryOperation UnaryArithmeticOperation Expression
	| BinaryOperation Expression BinaryArithmeticOperation Expression
	deriving (Eq, Ord)
instance Show Expression where
	show (Constant var) = show var
	show (UnaryOperation operation expr) = "(" ++ show operation ++ " " ++ show expr ++ ")"
	show (BinaryOperation expr1 operation expr2) = "(" ++ show expr1 ++ " " ++ show operation ++ " " ++ show expr2 ++ ")"
data ComparisonOperation =
	Equals | NotEquals | LessThan | GreaterThan
	deriving (Eq, Ord)
instance Show ComparisonOperation where
	show Equals = "=="
	show NotEquals = "!="
	show LessThan = "<"
	show GreaterThan = ">"
data UnaryArithmeticOperation =
	Count | Sum | Product
	deriving (Eq, Ord, Show)
data BinaryArithmeticOperation =
	Add | Subtract | Multiply | Divide
	deriving (Eq, Ord)
instance Show BinaryArithmeticOperation where
	show Add = "+"
	show Subtract = "-"
	show Multiply = "*"
	show Divide = "/"

--
-- Fact definitions.
--

-- A fact is a tuple of its name, the MAC, the time it was last modified, two
-- boolean flags (one is the "modified" flag, the other for internal use of the
-- simulation), and a list of the propterties of the fact.
data Fact = Fact Name MAC Time Bool Bool [Property]
instance Eq Fact where
	(==) (Fact name1 owner1 time1 modified1 preModified1 properties1) (Fact name2 owner2 time2 modified2 preModified2 properties2) =
		(name1 == name2) && (owner1 == owner2) && (time1 == time2) && (modified1 == modified2) && (preModified1 == preModified2) && (properties1 == properties2)
instance Show Fact where
	show (Fact name owner time modified preModified properties) =
		modifiedString1 ++ show time ++ modifiedString2 ++ " \t" ++ show owner ++ "." ++ name ++ propertiesString
		where	modifiedString1 = if modified then "[" else "("
			modifiedString2 = if modified then "]" else ")"
			propertiesString =	if (null properties) then ""
						else ": " ++ (join ", " (map show properties))
data Property = Property Key Variable
	deriving (Eq)
instance Show Property where
	show (Property key var) = key ++ " = " ++ show var



--
-- Functions definitions.
--

type Parameter = Type
data Function = Function Name [Parameter]
	deriving (Eq)
instance Show Function where
	show (Function name parameters) = name ++ "(" ++ (join ", " (map show parameters)) ++ ")"



--
-- Constants.
--

-- The broadcast MAC addresses all sensor nodes in the network.
broadcastMAC :: MAC
broadcastMAC = 0

-- Use as delimiter for concatenating strings in names and fact IDs.
separator :: String
separator = "."

-- Lower and upper bound for the priority.
minPriority :: Priority
minPriority = 0

maxPriority :: Priority
maxPriority = 100

-- Key strings of the predefined properties.
propertyKeyID :: String
propertyKeyID = "id"

propertyKeyOwner :: String
propertyKeyOwner = "owner"

propertyKeyTime :: String
propertyKeyTime = "time"

propertyKeyModified :: String
propertyKeyModified = "modified"

-- Updating a fact advances the simulated time by one unit, each simulation
-- step has room for 1000 of these operations.
simulationStepTime :: Time
simulationStepTime = 1000

factStepTime :: Time
factStepTime = 1

-- These booleans control the verbosity of the output of the simulation.
showRules :: Bool
showRules = False

showFacts :: Bool
showFacts = True

showFunctions :: Bool
showFunctions = False



--
-- Constructor functions.
--

varBool :: Bool -> Variable
varBool value = Variable TypeBool (ValueBool value)

varInt :: Int -> Variable
varInt value = Variable TypeInt (ValueInt value)

varFloat :: Float -> Variable
varFloat value = Variable TypeFloat (ValueFloat value)

varString :: String -> Variable
varString value = Variable TypeString (ValueString value)

varSlot :: Slot -> Variable
varSlot value = Variable TypeSlot (ValueSlot value)

newSimulation :: Network -> [Event] -> Simulation
newSimulation network events = Simulation (State 0 0 network) events

newEvent :: Int -> MAC -> [Statement] -> Event
newEvent step mac statements = Event step mac statements 

newNetwork :: Network
newNetwork = Network []

newNode :: MAC -> Position -> Node
newNode mac position
	| mac == broadcastMAC	= error ("Node MAC " ++ (show broadcastMAC) ++ " is reserved for broadcast packets.")
	| otherwise		= Node mac position [] [] []

newName :: Name -> Name -> Name
newName namespace name = namespace ++ separator ++ name

newRuleset :: Name -> [Rule] -> [Fact] -> [Function] -> Ruleset
newRuleset namespace rules facts functions =
	Ruleset (map (addRuleName namespace) rules) (map (addFactName namespace) facts) (map (addFunctionName namespace) functions)
	where	addRuleName namespace (Rule name priority conditions statements) =
			Rule (namespace ++ separator ++ name) priority conditions statements
		addFactName namespace (Fact name owner time modified preModified properties) =
			Fact (namespace ++ separator ++ name) owner time modified preModified properties
		addFunctionName namespace (Function name parameters) =
			Function (namespace ++ separator ++ name) parameters

newSlot :: Pattern -> Key -> [Condition] -> Slot
newSlot pattern key conditions = (Slot pattern key conditions)

-- A simple slot is just a short-hand for creating a slot that only refers to a
-- fact by the fact name.
newSimpleSlot :: Pattern -> Slot
newSimpleSlot pattern = (Slot pattern "" [])

newRule :: Name -> Priority -> [Condition] -> [Statement] -> Rule
newRule name priority conditions statements
	| (priority >= minPriority) && (priority <= maxPriority)	= Rule name priority conditions statements
	| otherwise	= error ("Invalid priority \"" ++ show priority ++ "\" for rule \"" ++ name ++ "\".")

newFact :: Name -> [Property] -> Fact
newFact name properties = Fact name 0 0 True False properties

newProperty :: Key -> Variable -> Property
newProperty key var
	| key == propertyKeyID		= error ("Property key \"" ++ propertyKeyID ++ "\" is reserved as fact identifier.")
	| key == propertyKeyOwner	= error ("Property key \"" ++ propertyKeyOwner ++ "\" is reserved as node identifier.")
	| key == propertyKeyTime	= error ("Property key \"" ++ propertyKeyTime ++ "\" is reserved as timestamp.")
	| key == propertyKeyModified	= error ("Property key \"" ++ propertyKeyModified ++ "\" is reserved as modified bit.")
	| otherwise			= Property key var

newFunction :: Name -> [Parameter] -> Function
newFunction name parameters = Function name parameters



--
-- Getters & Setters functions.
--

-- Add a list of nodes to a network.
addNodes :: Network -> [Node] -> Network
addNodes = foldl addNode

-- Add a sensor node to a network.
addNode :: Network -> Node -> Network
addNode (Network nodes) node
	| existsNode (Network nodes) (getMAC node)	= error "Node MAC already exists in this network."
	| otherwise					= Network (node:nodes)

-- Return to MAC of a node.
getMAC :: Node -> MAC
getMAC (Node mac _ _ _ _) = mac

-- Return the MACs of all nodes contained in a network.
getMACs :: Network -> [MAC]
getMACs (Network nodes) = map getMAC nodes

-- Check whether a node with a particular MAC exists in a network.
existsNode :: Network -> MAC -> Bool
existsNode network mac = any (mac ==) (getMACs network)

-- Return a node with a specified MAC from the network, if available.
getNode :: Network -> MAC -> Maybe Node
getNode (Network []) _ = Nothing
getNode (Network (node:nodes)) searchMAC
	| (getMAC node) == searchMAC	= Just node
	| otherwise			= getNode (Network nodes) searchMAC

-- Check whether a node has a specified MAC.
hasMAC :: MAC -> Node -> Bool
hasMAC searchMAC (Node mac _ _ _ _) = (searchMAC == mac)

-- Add a list of ruleset to a node.
addRulesets :: Node -> [Ruleset] -> Node
addRulesets = foldl addRuleset

-- Add a ruleset to a node.
addRuleset :: Node -> Ruleset -> Node
addRuleset node (Ruleset rules facts functions) = addFunctions (addFacts (addRules node rules) facts) functions

-- Add a list of rules to a node.
addRules :: Node -> [Rule] -> Node
addRules = foldl addRule

-- Add a rule to a node.
addRule :: Node -> Rule -> Node
addRule (Node mac position rules facts functions) rule = Node mac position newRules facts functions
	where	newRules = addPrioritizedRule rules rule
		addPrioritizedRule [] newRule = [newRule]
		addPrioritizedRule (rule:rules) newRule
			| priority >= newPriority	= rule:(addPrioritizedRule rules newRule)
			| otherwise			= newRule:rule:rules
			where	(Rule _ priority _ _) = rule
				(Rule _ newPriority _ _) = newRule

-- Add a list of facts to a node.
addFacts :: Node -> [Fact] -> Node
addFacts = foldl addFact

-- Add a fact to a node. If the fact doesn't have an owner, set to current
-- node.
addFact :: Node -> Fact -> Node
addFact node fact
	| owner == broadcastMAC	= addFact node (Fact name mac time modified preModified properties)
	| otherwise		= Node mac position rules (fact:facts) functions
	where	(Node mac position rules facts functions) = node
		(Fact name owner time modified preModified properties) = fact

-- Check whether a node contains a specified fact.
existsFact :: Node -> Pattern -> Bool
existsFact (Node mac position rules facts functions) pattern = any (matchesFact pattern) facts

-- Returns a particular fact from a node, if available.
getFact :: Node -> Pattern -> Maybe Fact
getFact (Node _ _ _ [] _) _ = Nothing
getFact (Node mac position rules (fact:facts) functions) pattern
	| name == pattern	= Just fact
	| otherwise		= getFact (Node mac position rules facts functions) pattern
	where (Fact name _ _ _ _ _) = fact

-- Remove a fact from a node.
removeFact :: Node -> Pattern -> [Condition] -> Node
removeFact (Node mac position rules facts functions) pattern conditions = Node mac position rules newFacts functions
	where	newFacts = filter (\fact -> (notElem fact removedFacts)) facts
		removedFacts = retrieveFacts facts pattern conditions

-- Add a list of functions to a node.
addFunctions :: Node -> [Function] -> Node
addFunctions = foldl addFunction

-- Add a function to a node.
addFunction :: Node -> Function -> Node
addFunction (Node mac position rules facts functions) function = Node mac position rules facts (function:functions)

-- Given a list of facts, e.g. the fact repository of a node, and a variable,
-- this function returns the variable if it's not a slot. For slots however, it
-- recursively resolves the slot by matching it against the given facts until
-- it resolves to one or many plain variables.
resolveVariable :: [Fact] -> Variable -> [Variable]
resolveVariable facts (Variable TypeSlot (ValueSlot slot))
	| null matchingFacts	= error ("Unable to resolve fact \"" ++ show slot ++ "\" when resolving variable.")
	| otherwise		= concat (map (resolveVariable facts) vars)
	where	Slot name key conditions = slot
		matchingFacts = retrieveFacts facts name conditions
		vars = map (_resolveVariable key) matchingFacts
		_resolveVariable key fact
			| existsVariable fact key	= getVariable fact key
			| otherwise			= error ("Unable to resolve key \"" ++ show slot ++ "\"  when resolving variable.")
			where (Fact name _ _ _ _ _) = fact
resolveVariable _ var = [var]

-- Given a pattern and a fact, this function checks whether the fact is
-- referenced by the pattern.
-- TODO: Maybe implement patterns with wildcards.
matchesFact :: Pattern -> Fact -> Bool
matchesFact pattern (Fact name _ _ _ _ _) = (pattern == name)

-- Given a list of facts (e.g. the fact repository of a node), a pattern and a
-- list of conditions (which not surprisingly is what makes up the addressing
-- capability of a slot), this function returns those facts that match both
-- pattern and conditions, i.e. it resolves a slot.
retrieveFacts :: [Fact] -> Pattern -> [Condition] -> [Fact]
retrieveFacts facts pattern conditions = filter (\fact -> and (map (checkCondition (fact:remainingFacts)) conditions)) candidateFacts
	where	candidateFacts = filter (matchesFact pattern) facts
		remainingFacts = filter (\fact -> notElem fact candidateFacts) facts

-- Check whether a fact has a property with one particular key.
existsVariable :: Fact -> Key -> Bool
existsVariable fact searchKey
	| searchKey == propertyKeyID		= True
	| searchKey == propertyKeyOwner		= True
	| searchKey == propertyKeyTime		= True
	| searchKey == propertyKeyModified	= True
	| otherwise				= any (searchKey ==) keys
	where	(Fact _ _ _ _ _ properties) = fact
		keys = map (\(Property key _) -> key) properties

-- Return the variable of a fact with a specified key.
getVariable :: Fact -> Key -> Variable
getVariable fact searchKey
	| not (existsVariable fact searchKey)	= error ("Can't retrieve var for key " ++ searchKey ++ ".")
	| searchKey == propertyKeyID		= Variable TypeString (ValueString (show owner ++ "." ++ show time))
	| searchKey == propertyKeyOwner		= Variable TypeInt (ValueInt owner)
	| searchKey == propertyKeyTime		= Variable TypeInt (ValueInt time)
	| searchKey == propertyKeyModified	= Variable TypeBool (ValueBool modified)
	| otherwise				= var
	where	(Property _ var) = head (filter (\(Property key _) -> (key == searchKey)) properties)
		(Fact _ owner time modified _ properties) = fact


--
-- Simulation functions.
--

-- This is the main simulation function. Based on a simulation object, it
-- generates the infinite list of all states that the simulated network will
-- eventually have.
runSimulation :: Simulation -> [State]
runSimulation (Simulation state events) = _runSimulation state
	where _runSimulation = iterate (processState events)

-- This function transform the current state of a simulated network to next
-- state. This involves several steps: 1) Process the events scheduled for this
-- simulation step. 2) Simulate the rule engines on all sensor nodes. 3) Mark
-- those facts as modified that have been updated during the simulation step.
-- 4) Adjust simulated time and simulation step count.
processState :: [Event] -> State -> State
processState events state
	| nextStepTime >= newTime	= (State (step + 1) nextStepTime (Network (map cleanNode newNodes)))
	| otherwise			= error (show (newTime - time) ++ " facts were altered in this step. Please increase simulationStepTime.")
	where	nextStepTime = time + simulationStepTime
		(State step time network) = state
		(State _ newTime (Network newNodes)) = processNetwork (processEvents state currentEvents) network
		currentEvents = filter (\(Event eventStep _ _) -> (eventStep == step)) events

-- Simulate how a list of events transforms the current state by executing the
-- statements that are contained in the events.
processEvents :: State -> [Event] -> State
processEvents state [] = state
processEvents state ((Event time mac statements):events)
	| mac == broadcastMAC		=
		foldl (\state mac -> processEvents state ((Event time mac statements):events)) state (getMACs network)
	| existsNode network mac	= processEvents (foldl (\state -> applyStatement state mac) state statements) events
	| otherwise			= error ("Can't apply events on node " ++ show mac ++ ".")
	where	(State _ _ network) = state
		Just (Node _ _ _ facts _) = getNode network mac

-- Simulate how a network is transformed from the current to the next state by
-- simulating all nodes of the network.
processNetwork :: State -> Network -> State
processNetwork state (Network nodes) = foldl processNode state nodes

-- Simulate how a node is transformed from the current to the next state by
-- applying its rules to the fact repository.
processNode :: State -> Node -> State
processNode state (Node mac _ rules _ _) = foldl (\state -> applyRule state mac) state rules

-- This function is equivalent to one run of the rule engine and as such
-- central to the FACTS system. It applies all statements of a rule to the fact
-- repository, if at least one fact referenced in the conditions of said rule
-- marked as modified and all conditions evaluate as true.
applyRule :: State -> MAC -> Rule -> State
applyRule state mac (Rule name _ conditions statements)
	| oneFactIsModified && allConditionsAreTrue	= foldl (\state -> applyStatement state mac) state statements
	| otherwise					= state
	where	oneFactIsModified	= or (map (checkFacts facts) conditions)
		allConditionsAreTrue	= and (map (checkCondition facts) conditions)
		facts = getFacts state mac

-- Given a list of facts (e.g. a fact repository) and a condition, this
-- function returns whether one of the facts mentioned in the condition is
-- marked as modified.
checkFacts :: [Fact] -> Condition -> Bool
checkFacts facts (Exists (Slot pattern _ conditions)) = isModified (retrieveFacts facts pattern conditions)
checkFacts facts (Eval expr1 _ expr2) = (_checkFacts facts expr1) || (_checkFacts facts expr2)
	where	_checkFacts facts (Constant (Variable TypeSlot (ValueSlot (Slot name _ _)))) = isModified (retrieveFacts facts name [])
		_checkFacts _ (Constant _) = False
		_checkFacts facts (UnaryOperation _ expr) = _checkFacts facts expr1
		_checkFacts facts (BinaryOperation expr1 _ expr2) = (_checkFacts facts expr1) || (_checkFacts facts expr2) 

-- Given a list of facts (e.g. a fact repository) and a condition, this
-- function returns whether the condition is true given these facts.
checkCondition :: [Fact] -> Condition -> Bool
checkCondition facts (Exists (Slot pattern _ conditions)) = not (null (retrieveFacts facts pattern conditions))
checkCondition facts (Eval expr1 operation expr2) = or (map (_checkCondition facts operation) (cross vars1 vars2))
	where	vars1 = evalExpression facts expr1
		vars2 = evalExpression facts expr2
		_checkCondition facts operation ((Variable type1 value1), (Variable type2 value2))
			| (type1 == type2)	= __checkCondition operation value1 value2
			| otherwise		= False
		__checkCondition Equals var1 var2 = (var1 == var2)
		__checkCondition NotEquals var1 var2 = (var1 /= var2)
		__checkCondition LessThan var1 var2 = (var1 < var2)
		__checkCondition GreaterThan var1 var2 = (var1 > var2)

-- Given a list of facts (e.g. a fact repository) and an expression, this
-- function evaluates the expression and returns the resulting values as
-- variables. Note that as the slots contained in an expression may address
-- multiple facts in the fact repository, the result of this function is a list
-- of resulting values derived from all possible combination (the cross
-- product) of matching facts.
evalExpression :: [Fact] -> Expression -> [Variable]
evalExpression facts (Constant var) = resolveVariable facts var
evalExpression facts (UnaryOperation Count expr) = [varInt (length (evalExpression facts expr))]
evalExpression facts (UnaryOperation Sum expr) = [foldl (\var1 var2 -> head (evalExpression facts (BinaryOperation (Constant var1) Add (Constant var2)))) (varInt 0) (evalExpression facts expr)]
evalExpression facts (UnaryOperation Product expr) = [foldl (\var1 var2 -> head (evalExpression facts (BinaryOperation (Constant var1) Multiply (Constant var2)))) (varInt 1) (evalExpression facts expr)]
evalExpression facts (BinaryOperation expr1 operation expr2) = map (_evalExpression operation) (cross vars1 vars2)
	where	vars1 = evalExpression facts expr1
		vars2 = evalExpression facts expr2
		_evalExpression Add ((Variable TypeBool (ValueBool value1)), (Variable TypeBool (ValueBool value2))) =
			Variable TypeBool (ValueBool (value1 || value2))
		_evalExpression Add ((Variable TypeInt (ValueInt value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeInt (ValueInt (value1 + value2))
		_evalExpression Add ((Variable TypeInt (ValueInt value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat ((fromRational (toRational value1)) + value2))
		_evalExpression Add ((Variable TypeFloat (ValueFloat value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeFloat (ValueFloat (value1 + (fromRational (toRational value2))))
		_evalExpression Add ((Variable TypeFloat (ValueFloat value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat (value1 + value2))
		_evalExpression Add ((Variable TypeString (ValueString value1)), (Variable TypeBool (ValueBool value2))) =
			Variable TypeString (ValueString (value1 ++ show value2))
		_evalExpression Add ((Variable TypeString (ValueString value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeString (ValueString (value1 ++ show value2))
		_evalExpression Add ((Variable TypeString (ValueString value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeString (ValueString (value1 ++ show value2))
		_evalExpression Add ((Variable TypeString (ValueString value1)), (Variable TypeString (ValueString value2))) =
			Variable TypeString (ValueString (value1 ++ value2))
		_evalExpression Subtract ((Variable TypeInt (ValueInt value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeInt (ValueInt (value1 - value2))
		_evalExpression Subtract ((Variable TypeInt (ValueInt value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat ((fromRational (toRational value1)) - value2))
		_evalExpression Subtract ((Variable TypeFloat (ValueFloat value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeFloat (ValueFloat (value1 - (fromRational (toRational value2))))
		_evalExpression Subtract ((Variable TypeFloat (ValueFloat value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat (value1 - value2))
		_evalExpression Multiply ((Variable TypeBool (ValueBool value1)), (Variable TypeBool (ValueBool value2))) =
			Variable TypeBool (ValueBool (value1 && value2))
		_evalExpression Multiply ((Variable TypeInt (ValueInt value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeInt (ValueInt (value1 * value2))
		_evalExpression Multiply ((Variable TypeInt (ValueInt value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat ((fromRational (toRational value1)) * value2))
		_evalExpression Multiply ((Variable TypeFloat (ValueFloat value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeFloat (ValueFloat (value1 * (fromRational (toRational value2))))
		_evalExpression Multiply ((Variable TypeFloat (ValueFloat value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat (value1 * value2))
		_evalExpression Divide ((Variable TypeInt (ValueInt value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeInt (ValueInt (value1 `div` value2))
		_evalExpression Divide ((Variable TypeInt (ValueInt value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat ((fromRational (toRational value1)) / value2))
		_evalExpression Divide ((Variable TypeFloat (ValueFloat value1)), (Variable TypeInt (ValueInt value2))) =
			Variable TypeFloat (ValueFloat (value1 / (fromRational (toRational value2))))
		_evalExpression Divide ((Variable TypeFloat (ValueFloat value1)), (Variable TypeFloat (ValueFloat value2))) =
			Variable TypeFloat (ValueFloat (value1 / value2))
		_evalExpression operation ((Variable type1 _), (Variable type2 _)) =
			error ("Operation " ++ show operation ++ " not implemented for types " ++ show type1 ++ " and " ++ show type2 ++ ".")

-- This function transforms the current state to the next state by applying a
-- statement to the fact repository on the node with the given MAC.
applyStatement :: State -> MAC -> Statement -> State
applyStatement state mac (Define name keyExpressionPairs) = addFactTo state mac fact
	where	fact = foldl (\fact (key, var) -> updateFact fact key var) (newFact name []) keyVariablePairs
		keyVariablePairs = concat (map (\(key, vars) -> [(key, var) | var <- vars]) keyVariablesPairs)
		keyVariablesPairs = map (\(key, expr) -> (key, (evalExpression facts expr))) keyExpressionPairs
		facts = getFacts state mac
applyStatement state mac (Retract (Slot pattern _ conditions)) = removeFactFrom state mac pattern conditions
applyStatement state mac (Send receiverMACExpression txPowerExpression (Slot pattern _ conditions)) =
	foldl (\state -> _applySendStatement state mac pattern conditions) state (cross receiverMACVars txPowerVars)
	where	receiverMACVars = evalExpression facts receiverMACExpression
		txPowerVars = evalExpression facts txPowerExpression
		facts = getFacts state mac
		_applySendStatement state mac pattern conditions (receiverMACVar, txPowerVar)
			| (receiverMACType == TypeInt) && (txPowerType == TypeFloat)	= sendFactTo state mac receiverMAC txPower pattern conditions
			| otherwise	= error "receiverMAC or txPower expression do not eval to proper type."
			where	(Variable receiverMACType receiverMACValue) = receiverMACVar
				(ValueInt receiverMAC) = receiverMACValue
				(Variable txPowerType txPowerValue) = txPowerVar
				(ValueFloat txPower) = txPowerValue
applyStatement state mac (Set slot expression) =
	_applySetStatement 0 iterations state mac slot expression
	where	iterations = length (evalExpression (getFacts state mac) expression)
		_applySetStatement counter iterations state mac slot expression
			| counter < iterations	= _applySetStatement (counter + 1) iterations (updateFactAt state mac slot ((evalExpression facts expression)!!counter)) mac slot expression
			| otherwise		= state
			where	facts = getFacts state mac
-- TODO: Implement calling functions.
applyStatement state mac (Call _ _) = state

-- This function returns the fact repository of a node in a given state of the
-- simulation.
getFacts :: State -> MAC -> [Fact]
getFacts (State step time network) mac
	| existsNode network mac	= facts
	| otherwise			= error ("Unable to get facts of node " ++ show mac ++ ".")
	where	Just (Node _ _ _ facts _) = getNode network mac

-- This function adds a new fact to a node in the simulation.
addFactTo :: State -> MAC -> Fact -> State
addFactTo (State step time (Network nodes)) mac fact
	| existsNode (Network nodes) mac	= State step newTime newNetwork
	| otherwise				= error ("Unable to add fact to node " ++ show mac ++ ".")
	where	newNetwork = Network (replace (hasMAC mac) newNode nodes)
		newNode = addFact node (touchFact newTime fact)
		newTime = time + factStepTime
		Just node = getNode (Network nodes) mac

-- This function removes a fact from a node in the simulation.
removeFactFrom :: State -> MAC -> Pattern -> [Condition] -> State
removeFactFrom (State step time (Network nodes)) mac pattern conditions
	| existsNode (Network nodes) mac	= State step newTime newNetwork
	| otherwise				= error ("Unable to remove fact from node " ++ show mac ++ ".")
	where	newNetwork = Network (replace (hasMAC mac) newNode nodes)
		newNode = removeFact node pattern conditions
		newTime = time + factStepTime
		Just node = getNode (Network nodes) mac

-- This function sends a set of facts as specified by a pattern and a list of
-- conditions, i.e. the addressing information found in a slot, from one node
-- to another one (or to many, if the receiver MAC is the broadcast MAC).
-- Whether a node actually receives the transmission is controlled by the
-- transmit range.
sendFactTo :: State -> MAC -> MAC -> Float -> Pattern -> [Condition] -> State
sendFactTo state senderMAC receiverMAC range pattern conditions
	| receiverMAC == broadcastMAC	=
		foldl (\state mac -> sendFactTo state senderMAC mac range pattern conditions) state (getMACs network)
	| senderMAC == receiverMAC	= state
	| sender == Nothing		= error ("Unable to find sending node " ++ show senderMAC ++ ".")
	| receiver == Nothing		= error ("Unable to find receiving node " ++ show receiverMAC ++ ".")
	| isReachable			= foldl (\state -> addFactTo state receiverMAC) state (retrieveFacts facts pattern conditions)
	| otherwise			= state
	where	(State _ _ network) = state
		sender = getNode network senderMAC
		receiver = getNode network receiverMAC
		Just (Node _ (x1, y1) _ facts _) = sender
		Just (Node _ (x2, y2) _ _ _) = receiver
		isReachable = (sqrt ((x1-x2)^2 + (y1-y2)^2)) <= range

-- This function updates one or many values of the property of those facts that
-- are resolved based on node MAC and a slot.
updateFactAt :: State -> MAC -> Slot -> Variable -> State
updateFactAt (State step time (Network nodes)) searchMAC slot var
	| (existsNode (Network nodes) searchMAC) && (existsFact node slotName)	= State step newTime newNetwork
	| otherwise	= error ("Unable to update fact " ++ slotName ++ " in node " ++ show mac ++ ".")
	where	Just node = getNode (Network nodes) searchMAC
		(Node mac position rules facts functions) = node
		(Slot slotName _ _) = slot
		newTime = time + factStepTime
		newNetwork = Network (replace (hasMAC searchMAC) newNode nodes)
		newNode = Node mac position rules newFacts functions
		newFacts = _updateFactAt facts slot var newTime
		_updateFactAt [] _ _ _ = []
		_updateFactAt (fact:facts) slot var time
			| (matchesFact slotName fact)	= newFact:(_updateFactAt facts slot var time)
			| otherwise			= fact:(_updateFactAt facts slot var time)
			where	(Slot slotName slotKey _) = slot
				newFact = touchFact newTime (updateFact fact slotKey var)

-- This function updates the property of a fact and returns the new fact.
updateFact :: Fact -> Key -> Variable -> Fact
updateFact (Fact name owner time modified preModified properties) key var = (Fact name owner time modified preModified newProperties)
	where	newProperties = updateProperties properties key var

-- This function updates the list of properties of a fact and returns the new
-- list.
updateProperties :: [Property] -> Key -> Variable -> [Property]
updateProperties [] key var = [newProperty key var]
updateProperties ((Property key var):properties) searchKey var2
	| (key == searchKey)	= (Property key (updateVariable var var2)):properties
	| otherwise		= (Property key var):(updateProperties properties searchKey var2)

-- This function updates a variable and returns it. If the types of the new and
-- the old variables don't match, processing aborts with an error.
updateVariable :: Variable -> Variable -> Variable
updateVariable (Variable type1 value1) (Variable type2 value2)
	| (type1 == type2)	= (Variable type2 value2)
	| otherwise		= error "Type mismatch while updating vars."

-- This function updates the modified information of all facts of a node.
cleanNode :: Node -> Node
cleanNode (Node mac position rules facts functions) = Node mac position rules (map cleanFact facts) functions

-- This function updates the modified information of a fact.
cleanFact :: Fact -> Fact
cleanFact (Fact name owner time modified preModified properties) = (Fact name owner time preModified False properties)

-- This function is to be called whenever a fact is modified. It updates the
-- time of last access and marks the fact as modified.
touchFact :: Time -> Fact -> Fact
touchFact time (Fact name owner _ modified preModified properties) = (Fact name owner time modified True properties)

-- This function checks whether one of a set of facts is flagged as modified.
isModified :: [Fact] -> Bool
isModified facts = any (\(Fact _ _ _ modified _ _) -> modified) facts



--
-- Utility functions.
--

-- Given two lists, this function returns the cross product of the two lists.
cross :: [a] -> [b] -> [(a,b)]
cross xs ys = [(x,y) | x <- xs, y <- ys]

-- This function returns the concatenation of a list of strings with delimiter
-- inserted in between.
join :: String -> [String] -> String
join _ [] = ""
join _ (s:[]) = s
join delim (s:ss) = s ++ delim ++ (join delim ss)

-- This function replaces all occurences of a particular object in a list as
-- specified by a matcher function with a new given object.
replace :: (a -> Bool) -> a -> [a] -> [a]
replace _ _ [] = []
replace f x (y:ys)
	| f y		= x:(replace f x ys)
	| otherwise	= y:(replace f x ys)
