--
-- This is the Turing Machine ruleset which implements a complete Turing
-- Machine to show off the power of the rule engine.
--

module FACTS.Rulesets.TuringMachine (

	ruleset,
	constErrorStateSlot, constBlankSymbolSlot, constRightSlot, constLeftSlot, constNeutralSlot,
	runName, functionName, tapeName

) where

import FACTS.Main

namespace = "TM"

ruleset = newRuleset namespace [leftBorderRule, rightBorderRule, stepRule, errorRule] [constantsFact] []

-- These two rules create the impression of an infinite tape by extending the
-- tape with new tape facts whenever the head reaches a side.
leftBorderRule = newRule "leftBorder" maxPriority [condition1] [statement1, statement2]
	where	condition1 = Eval (Constant (varSlot runPositionSlot)) LessThan (Constant (varSlot runLeftBorderSlot))
		statement1 = Define tapeName [
			("position", Constant (varSlot runPositionSlot)),
			("symbol", Constant (varSlot constBlankSymbolSlot))]
		statement2 = Set runLeftBorderSlot (BinaryOperation (Constant (varSlot runLeftBorderSlot)) Subtract (Constant (varInt 1)))

rightBorderRule = newRule "rightBorder" maxPriority [condition1] [statement1, statement2]
	where	condition1 = Eval (Constant (varSlot runPositionSlot)) GreaterThan (Constant (varSlot runRightBorderSlot))
		statement1 = Define tapeName [
			("position", Constant (varSlot runPositionSlot)),
			("symbol", Constant (varSlot constBlankSymbolSlot))]
		statement2 = Set runRightBorderSlot (BinaryOperation (Constant (varSlot runRightBorderSlot)) Add (Constant (varInt 1)))

-- This rule does the main processing of the Turing Machine: As long as the
-- machine is running, move the current state, position and symbol to a
-- temporary fact, then update the tape fact, and the state and position
-- attributes of the run fact.
stepRule = newRule "step" minPriority [condition1] [statement1, statement2, statement3, statement4, statement5, statement6]
	where	condition1 = Exists (newSimpleSlot runName)
		statement1 = Define currentName [
			("state", Constant (varSlot runStateSlot)),
			("position", Constant (varSlot runPositionSlot)),
			("symbol", Constant (varSlot tapeSymbolSlot))]
		statement2 = Retract (newSlot tapeName "" [Eval (Constant (varSlot tapePositionSlot)) Equals (Constant (varSlot runPositionSlot))])
		statement3 = Define tapeName [
			("position", Constant (varSlot runPositionSlot)),
			("symbol", Constant (varSlot functionNextSymbolSlot))]
		statement4 = Set runStateSlot (Constant (varSlot functionNextStateSlot))
		statement5 = Set runPositionSlot (BinaryOperation (Constant (varSlot currentPositionSlot)) Add (Constant (varSlot functionNextMovementSlot)))
		statement6 = Retract (newSimpleSlot currentName)

-- This rule deals with error handling and turns the machine off when an error
-- is encountered.
errorRule = newRule "error" maxPriority [condition1] [statement1]
	where	condition1 = Eval (Constant (varSlot runStateSlot)) Equals (Constant (varSlot constErrorStateSlot))
		statement1 = Retract (newSimpleSlot runName)

-- These constants are to be used by external scripts that use this ruleset.
constantsFact = newFact "constants" [
	newProperty "errorState" (varInt (-1)),
	newProperty "blankSymbol" (varString "_"),
	newProperty "right" (varInt 1),
	newProperty "left" (varInt (-1)),
	newProperty "neutral" (varInt 0)]

constantsName = newName namespace "constants"
runName = newName namespace "run"
currentName = newName namespace "current"
functionName = newName namespace "function"
tapeName = newName namespace "tape"

-- These slots make the programming above more readable.
constErrorStateSlot = newSlot constantsName "errorState" []
constBlankSymbolSlot = newSlot constantsName "blankSymbol" []
constRightSlot = newSlot constantsName "right" []
constLeftSlot = newSlot constantsName "left" []
constNeutralSlot = newSlot constantsName "neutral" []

runStateSlot = newSlot runName "state" []
runPositionSlot = newSlot runName "position" []
runLeftBorderSlot = newSlot runName "leftBorder" []
runRightBorderSlot = newSlot runName "rightBorder" []

currentStateSlot = newSlot currentName "state" []
currentPositionSlot = newSlot currentName "position" []
currentSymbolSlot = newSlot currentName "symbol" []

-- Note how the slots for next state, symbol, and movement filter the fact base
-- dynamically based on the current state of the Turing Machine.
functionStateSlot = newSlot functionName "state" []
functionSymbolSlot = newSlot functionName "symbol" []
functionNextStateSlot = newSlot functionName "nextState" [
	Eval (Constant (varSlot functionStateSlot)) Equals (Constant (varSlot currentStateSlot)),
	Eval (Constant (varSlot functionSymbolSlot)) Equals (Constant (varSlot currentSymbolSlot))]
functionNextSymbolSlot = newSlot functionName "nextSymbol" [
	Eval (Constant (varSlot functionStateSlot)) Equals (Constant (varSlot currentStateSlot)),
	Eval (Constant (varSlot functionSymbolSlot)) Equals (Constant (varSlot currentSymbolSlot))]
functionNextMovementSlot = newSlot functionName "nextMovement" [
	Eval (Constant (varSlot functionStateSlot)) Equals (Constant (varSlot currentStateSlot)),
	Eval (Constant (varSlot functionSymbolSlot)) Equals (Constant (varSlot currentSymbolSlot))]

tapePositionSlot = newSlot tapeName "position" []
tapeSymbolSlot = newSlot tapeName "symbol" [
	Eval (Constant (varSlot tapePositionSlot)) Equals (Constant (varSlot runPositionSlot))]
