--
-- This script implements a Universal Turing Machine.
--
-- See Yurii Rogozhin. Small Universal Turing Machines. Theoretical Computer
-- Science, 168(2):215-240, November 1996.
--
-- This implements UTM(4,6) from page 231.
--

import FACTS.Main
import qualified FACTS.Rulesets.TuringMachine as TM

-- Encode state and tape symbols just as in the paper.
q1 = varInt 1
q2 = varInt 2
q3 = varInt 3
q4 = varInt 4
qE = (varSlot TM.constErrorStateSlot)

s0 = (varSlot TM.constBlankSymbolSlot)
s1 = varString "1"
sB = varString "B"
sBl = varString "Bl"
sBr = varString "Br"
sC = varString "C"

r = (varSlot TM.constRightSlot)
l = (varSlot TM.constLeftSlot)
n = (varSlot TM.constNeutralSlot)

-- These facts implement the function of UTM(4,6).
functionFact1 = newFact TM.functionName [newProperty "state" q1, newProperty "symbol" s1, newProperty "nextState" q1, newProperty "nextSymbol" sBl, newProperty "nextMovement" l]
functionFact2 = newFact TM.functionName [newProperty "state" q1, newProperty "symbol" sB, newProperty "nextState" q1, newProperty "nextSymbol" sBr, newProperty "nextMovement" r]
functionFact3 = newFact TM.functionName [newProperty "state" q1, newProperty "symbol" sBr, newProperty "nextState" q1, newProperty "nextSymbol" sB, newProperty "nextMovement" l]
functionFact4 = newFact TM.functionName [newProperty "state" q1, newProperty "symbol" sBl, newProperty "nextState" q1, newProperty "nextSymbol" s0, newProperty "nextMovement" r]
functionFact5 = newFact TM.functionName [newProperty "state" q1, newProperty "symbol" s0, newProperty "nextState" q1, newProperty "nextSymbol" sBl, newProperty "nextMovement" l]
functionFact6 = newFact TM.functionName [newProperty "state" q1, newProperty "symbol" sC, newProperty "nextState" q4, newProperty "nextSymbol" s0, newProperty "nextMovement" r]
functionFact7 = newFact TM.functionName [newProperty "state" q2, newProperty "symbol" s1, newProperty "nextState" q2, newProperty "nextSymbol" s0, newProperty "nextMovement" r]
functionFact8 = newFact TM.functionName [newProperty "state" q2, newProperty "symbol" sB, newProperty "nextState" q3, newProperty "nextSymbol" sBr, newProperty "nextMovement" l]
functionFact9 = newFact TM.functionName [newProperty "state" q2, newProperty "symbol" sBr, newProperty "nextState" q2, newProperty "nextSymbol" sBl, newProperty "nextMovement" r]
functionFact10 = newFact TM.functionName [newProperty "state" q2, newProperty "symbol" sBl, newProperty "nextState" q2, newProperty "nextSymbol" sBr, newProperty "nextMovement" l]
functionFact11 = newFact TM.functionName [newProperty "state" q2, newProperty "symbol" s0, newProperty "nextState" q2, newProperty "nextSymbol" s1, newProperty "nextMovement" l]
functionFact12 = newFact TM.functionName [newProperty "state" q2, newProperty "symbol" sC, newProperty "nextState" q2, newProperty "nextSymbol" sB, newProperty "nextMovement" r]
functionFact13 = newFact TM.functionName [newProperty "state" q3, newProperty "symbol" s1, newProperty "nextState" q3, newProperty "nextSymbol" s1, newProperty "nextMovement" r]
functionFact14 = newFact TM.functionName [newProperty "state" q3, newProperty "symbol" sB, newProperty "nextState" q4, newProperty "nextSymbol" sBl, newProperty "nextMovement" r]
functionFact15 = newFact TM.functionName [newProperty "state" q3, newProperty "symbol" sBr, newProperty "nextState" q3, newProperty "nextSymbol" sB, newProperty "nextMovement" r]
functionFact16 = newFact TM.functionName [newProperty "state" q3, newProperty "symbol" sBl, newProperty "nextState" qE, newProperty "nextSymbol" s0, newProperty "nextMovement" n]
functionFact17 = newFact TM.functionName [newProperty "state" q3, newProperty "symbol" s0, newProperty "nextState" q1, newProperty "nextSymbol" sC, newProperty "nextMovement" r]
functionFact18 = newFact TM.functionName [newProperty "state" q3, newProperty "symbol" sC, newProperty "nextState" q1, newProperty "nextSymbol" s1, newProperty "nextMovement" r]
functionFact19 = newFact TM.functionName [newProperty "state" q4, newProperty "symbol" s1, newProperty "nextState" q4, newProperty "nextSymbol" s0, newProperty "nextMovement" r]
functionFact20 = newFact TM.functionName [newProperty "state" q4, newProperty "symbol" sB, newProperty "nextState" q2, newProperty "nextSymbol" sC, newProperty "nextMovement" l]
functionFact21 = newFact TM.functionName [newProperty "state" q4, newProperty "symbol" sBr, newProperty "nextState" q4, newProperty "nextSymbol" sBl, newProperty "nextMovement" r]
functionFact22 = newFact TM.functionName [newProperty "state" q4, newProperty "symbol" sBl, newProperty "nextState" qE, newProperty "nextSymbol" s0, newProperty "nextMovement" n]
functionFact23 = newFact TM.functionName [newProperty "state" q4, newProperty "symbol" s0, newProperty "nextState" q2, newProperty "nextSymbol" sC, newProperty "nextMovement" l]
functionFact24 = newFact TM.functionName [newProperty "state" q4, newProperty "symbol" sC, newProperty "nextState" q4, newProperty "nextSymbol" sB, newProperty "nextMovement" r]

-- This implements the example from page 220.
-- TODO: This doesn't actually work. I'm unsure on how to encode the input.
tapeFact1 = newFact TM.tapeName [newProperty "position" (varInt 1), newProperty "symbol" s1]
tapeFact2 = newFact TM.tapeName [newProperty "position" (varInt 2), newProperty "symbol" s1]
tapeFact3 = newFact TM.tapeName [newProperty "position" (varInt 3), newProperty "symbol" s1]

-- Initialize node with Turing Mashine ruleset, function facts, and tape facts.
node1 = addRulesets (newNode 1 (0,0)) [TM.ruleset]
node1f = addFacts node1 [functionFact1, functionFact2, functionFact3, functionFact4, functionFact5, functionFact6, functionFact7, functionFact8, functionFact9, functionFact10, functionFact11, functionFact12, functionFact13, functionFact14, functionFact15, functionFact16, functionFact17, functionFact18, functionFact19, functionFact20, functionFact21, functionFact22, functionFact23, functionFact24]
node1fb = addFacts node1f [tapeFact1, tapeFact2, tapeFact3]

-- Add node to the network.
network = addNodes newNetwork [node1fb]

-- Create an event that defines the initial state of the simulation and hence
-- of the Turing Machine.
startEvent = newEvent 0 0 [
	(Define TM.runName [
		("state", Constant q1),
		("position", Constant (varInt 1)),
		("leftBorder", Constant (varInt 1)),
		("rightBorder", Constant (varInt 3))])]
events = [startEvent]

-- Initialize simulation and display results.
sim :: Int -> IO()
sim state = putStr(show ((runSimulation (newSimulation network events))!!state))
